@@ -542,8 +542,17 @@ input_available (gpointer data, gint source, GdkInputCondition cond)
 		display_image (p);
 	      else
 		{
-		  gdk_input_remove (p->input_tag);
-		  p->input_tag = -1;
+		  if (p->input_tag < 0)
+		    {
+		      display_maybe (p);
+		      while (gtk_events_pending ())
+			gtk_main_iteration ();
+		    }
+		  else
+		    {
+		      gdk_input_remove (p->input_tag);
+		      p->input_tag = -1;
+		    }
 		  scan_start (p);
 		  break;
 		}
@@ -557,24 +566,82 @@ input_available (gpointer data, gint source, GdkInputCondition cond)
 	  scan_done (p);
 	  return;
 	}
-      if (!len)
-	break;			/* out of data for now */
+      if (!len) /* out of data for now */
+	{
+	  if (p->input_tag < 0)
+	    {
+	      display_maybe (p);
+	      while (gtk_events_pending ())
+		gtk_main_iteration ();
+	      continue;
+	    }
+	  else
+	    break;
+	}
 
       switch (p->params.format)
 	{
 	case SANE_FRAME_RGB:
-	  if (p->params.depth != 8)
-	    goto bad_depth;
-
-	  for (i = 0; i < len; ++i)
+	  switch (p->params.depth)
 	    {
-	      p->image_data[p->image_offset++] = buf[i];
-	      if (p->image_offset%3 == 0)
+	    case 1:
+	      for (i = 0; i < len; ++i)
 		{
-		  if (++p->image_x >= p->image_width
-		      && increment_image_y (p) < 0)
-		    return;
+		  u_char mask = buf[i];
+		  
+		  for (j = 0; j < 8; ++j)
+		    {
+		      u_char gl = (mask & (1 << j)) ? 0xff : 0x00;
+		      p->image_data[p->image_offset] = gl;
+		      if (j < 7)
+			p->image_offset += 3;
+		      else
+			{
+			  if ((p->image_offset % 3) != 2)
+			    p->image_offset -= 20;
+			  else
+			    p->image_offset++;
+			}
+
+		      if ((i % 3) == 0 && ++p->image_x >= p->image_width)
+			{
+			  if (increment_image_y (p) < 0)
+			    return;
+			  break;	/* skip padding bits */
+			}
+		    }
 		}
+	      
+	      break;
+
+	    case 8:
+	      for (i = 0; i < len; ++i)
+		{
+		  p->image_data[p->image_offset++] = buf[i];
+		  if (p->image_offset%3 == 0)
+		    {
+		      if (++p->image_x >= p->image_width
+			  && increment_image_y (p) < 0)
+			return;
+		    }
+		}
+	      break;
+	    case 16:
+	      for (i = 0; i < len; ++i)
+		{
+		  guint16 value = buf [i];
+		  if ((i % 2) == 1)
+		    p->image_data[p->image_offset++] = *(guint8 *) (&value);
+		  if (p->image_offset%6 == 0)
+		    {
+		      if (++p->image_x >= p->image_width
+			  && increment_image_y (p) < 0)
+			return;
+		    }
+		}
+	      break;
+	    default:
+	      goto bad_depth;
 	    }
 	  break;
 
@@ -614,6 +681,27 @@ input_available (gpointer data, gint source, GdkInputCondition cond)
 		    return;
 		}
 	      break;
+	    case 16:
+	      for (i = 0; i < len; ++i)
+		{
+		  guint16 value = buf [i];
+		  if ((i % 2) == 1)
+		    {
+		      p->image_data[p->image_offset++] = 
+			*(guint8 *) (&value);
+		      p->image_data[p->image_offset++] = 
+			*(guint8 *) (&value);
+		      p->image_data[p->image_offset++] = 
+			*(guint8 *) (&value);
+		    }
+		  if (p->image_offset%2 == 0)
+		    {
+		      if (++p->image_x >= p->image_width
+			  && increment_image_y (p) < 0)
+			return;
+		    }
+		}
+	      break;
 
 	    default:
 	      goto bad_depth;
@@ -634,7 +722,7 @@ input_available (gpointer data, gint source, GdkInputCondition cond)
 		    {
 		      u_char gl = (mask & 1) ? 0xff : 0x00;
 		      mask >>= 1;
-		      p->image_data[p->image_offset++] = gl;
+		      p->image_data[p->image_offset] = gl;
 		      p->image_offset += 3;
 		      if (++p->image_x >= p->image_width
 			  && increment_image_y (p) < 0)
@@ -653,6 +741,24 @@ input_available (gpointer data, gint source, GdkInputCondition cond)
 		    return;
 		}
 	      break;
+	      
+	    case 16:
+	      for (i = 0; i < len; ++i)
+		{
+		  guint16 value = buf [i];
+		  if ((i % 2) == 1)
+		    {
+		      p->image_data[p->image_offset] =  *(guint8 *) (&value);
+		      p->image_offset += 3;
+		    }
+		  if (p->image_offset % 2 == 0)
+		    {
+		      if (++p->image_x >= p->image_width
+			  && increment_image_y (p) < 0)
+			return;
+		    }
+		}
+	      break;
 
 	    default:
 	      goto bad_depth;
