@@ -602,8 +602,16 @@ input_available (gpointer data, gint source, GdkInputCondition cond)
 	    {
 	      if (!scan_win.param.last_frame)
 		{
-		  gdk_input_remove (scan_win.input_tag);
-		  scan_win.input_tag = -1;
+		  if (scan_win.input_tag < 0)
+		    {
+		      while (gtk_events_pending ())
+			gtk_main_iteration ();
+		    }
+		  else
+		    {
+		      gdk_input_remove (scan_win.input_tag);
+		      scan_win.input_tag = -1;
+		    }
 		  scan_start ();
 		  break;
 		}
@@ -617,8 +625,18 @@ input_available (gpointer data, gint source, GdkInputCondition cond)
 	  scan_done ();
 	  return;
 	}
-      if (!len)
-	break;			/* out of data for now */
+      if (!len) /* out of data for now */
+	{
+	  if (scan_win.input_tag < 0)
+	    {
+	      while (gtk_events_pending ())
+		gtk_main_iteration ();
+	      continue;
+	    }
+	  else
+	    break;
+	}
+
       scan_win.bytes_read += len;
       progress_update (scan_win.progress,
 		       scan_win.bytes_read / (gfloat) scan_win.num_bytes);
